# flake8: noqa=W293
# pylint: disable=line-too-long,missing-class-docstring,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name

"""
Test script for opening a youtube channel and getting the latest videos.
"""

import concurrent.futures
import json
import os
import time
import traceback
import unicodedata
import warnings
from dataclasses import dataclass
from typing import Any, Callable, Generator

import requests
from bs4 import BeautifulSoup  # type: ignore
from open_webdriver import open_webdriver  # type: ignore
from selenium.common.exceptions import (
    StaleElementReferenceException as StaleElementException,
)

IS_GITHUB_RUNNER = os.environ.get("GITHUB_ACTIONS") == "true"
# Always headless now.
HEADLESS = IS_GITHUB_RUNNER or True

URL = "https://www.youtube.com/@silverguru/videos"

JS_SCROLL_TO_BOTTOM = "window.scrollTo(0, document.documentElement.scrollHeight);"
JS_SCROLL_TO_BOTTOM_WAIT = 1
URL_BASE = "https://www.youtube.com"

CACHE_OUTER_HTML: dict[str, str] = {}


@dataclass
class YtVid:
    url: str
    title: str

    # needed for set membership
    def __hash__(self):
        return hash(self.url)

    def __eq__(self, other):
        return self.url == other.url

    def __repr__(self) -> str:
        data = self.to_dict()
        return json.dumps(data)

    def to_dict(self) -> dict:
        return {"url": self.url, "title": self.title}


def sanitize_filepath(path: str, replacement_char: str = "_") -> str:
    """
    Sanitize a file path to make it safe for use in file systems.

    :param path: The file path to be sanitized.
    :param replacement_char: The character to use for replacing invalid characters.
    :return: A sanitized version of the file path.
    """
    # Normalize unicode characters
    path = unicodedata.normalize("NFKD", path).encode("ascii", "ignore").decode("ascii")

    # Replace invalid file path characters
    invalid_chars = r'<>:"/\\|?*'
    for char in invalid_chars:
        path = path.replace(char, replacement_char)

    # Replace leading/trailing periods and spaces
    path = path.strip(". ")

    # Avoid reserved names in Windows like CON, PRN, AUX, NUL, etc.
    reserved_names = ["CON", "PRN", "AUX", "NUL"] + [f"{name}{i}" for name in ["COM", "LPT"] for i in range(1, 10)]
    basename = path.split("/")[-1]
    if basename.upper() in reserved_names:
        path = path.replace(basename, replacement_char + basename)

    # Truncate long paths
    max_length = 255
    if len(path) > max_length:
        extension = "." + path.split(".")[-1] if "." in path else ""
        path = path[: max_length - len(extension)] + extension
    path = path.replace("'", "_")

    return path


def parse_youtube_videos(div_strs: list[str]) -> list[YtVid]:
    """Div containing the youtube video, which has a title and an href."""
    out: list[YtVid] = []
    for div_str in div_strs:
        soup = BeautifulSoup(div_str, "html.parser")
        title_link = soup.find("a", id="video-title-link")
        try:
            title = title_link.get("title")
            href = title_link.get("href")
            assert title is not None
            assert href is not None
            assert href.startswith("/")
            href = URL_BASE + href
            title = sanitize_filepath(title.strip())
        except KeyboardInterrupt:
            return out
        except SystemExit:
            return out
        except Exception as err:  # pylint: disable=broad-except
            stack_trace = traceback.format_exc()
            warnings.warn(f"Error, could not scrape video: {err} {stack_trace}")
            continue
        out.append(YtVid(title=title, url=href))
    return out


def fetch_source_cached(unique_id: str, value_cb: Callable[[], str]) -> tuple[str, bool]:
    if unique_id in CACHE_OUTER_HTML:
        return CACHE_OUTER_HTML[unique_id], True
    value = value_cb()
    CACHE_OUTER_HTML[unique_id] = value
    return value, False


def clear_source_cache() -> None:
    CACHE_OUTER_HTML.clear()


def fetch_all_sources(yt_channel_url: str, limit: int = -1) -> Generator[str, None, None]:
    clear_source_cache()
    max_index = limit if limit > 0 else 1000
    with open_webdriver(headless=HEADLESS) as driver:

        def get_vid_attribute(vid: Any) -> str | None:
            try:
                try:
                    unique_id = getattr(vid, "id")
                    html, cached = fetch_source_cached(unique_id, lambda: str(vid.get_attribute("outerHTML")))
                    if cached:
                        return None
                    return html
                except AttributeError:
                    pass
                return str(vid.get_attribute("outerHTML"))
            except StaleElementException:
                warnings.warn("skipping stale element")
                return None

        def get_contents() -> list[str]:
            vids = driver.find_elements_by_tag_name("ytd-rich-item-renderer")
            with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
                futures: list = []
                for vid in vids:
                    # Only new videos will be resolved. All cached values will be None.
                    future = executor.submit(get_vid_attribute, vid)
                    futures.append(future.result)
                out = [future() for future in futures]
                # filter out None and empty strings (happens if error).
                out = [x for x in out if x]
            return out

        # All Chromium / web driver dependencies are now installed.
        driver.get(yt_channel_url)
        time.sleep(JS_SCROLL_TO_BOTTOM_WAIT)
        for item in get_contents():
            yield item
        last_scroll_height = 0
        for index in range(max_index + 1):
            driver.execute_script(JS_SCROLL_TO_BOTTOM)
            time.sleep(JS_SCROLL_TO_BOTTOM_WAIT)
            # yield get_contents()
            for item in get_contents():
                yield item
            scroll_height = driver.execute_script("return document.documentElement.scrollHeight")
            print(f"#### {index}: scrolling for new content ####")
            scroll_diff = abs(scroll_height - last_scroll_height)
            if scroll_diff < 100:
                break
            last_scroll_height = scroll_height
        if index == max_index and limit <= 0:
            warnings.warn("Reached max scroll limit.")


def test_channel_url(channel_url: str) -> bool:
    """Test if the channel url is valid."""
    response = requests.get(channel_url, timeout=10)
    return response.status_code == 200


def fetch_all_vids(yt_channel_url: str, limit: int = -1) -> list[YtVid]:
    """
    Open a web driver and navigate to Google. yt_channel_url should be
    of the form https://www.youtube.com/@silverguru/videos
    """
    if not test_channel_url(yt_channel_url):
        raise ValueError(f"Invalid channel url: {yt_channel_url}")
    pending_fetches = fetch_all_sources(yt_channel_url=yt_channel_url, limit=limit)
    list_vids: list[list[YtVid]] = []
    num_workers = max(1, os.cpu_count() or 0)
    with concurrent.futures.ThreadPoolExecutor(max_workers=num_workers) as executor:
        future_to_vid = {}
        for sources in pending_fetches:
            future = executor.submit(parse_youtube_videos, [sources])
            future_to_vid[future] = sources
        for future in concurrent.futures.as_completed(future_to_vid):
            vids = future.result()
            list_vids.append(vids)
    unique_vids: set[YtVid] = set()
    out_vids: list[YtVid] = []
    for vids in list_vids:
        for vid in vids:
            if vid not in unique_vids:
                unique_vids.add(vid)
                out_vids.append(vid)
    return out_vids


def main() -> int:
    vidlist = fetch_all_vids(URL, limit=1)
    print(f"Found {len(vidlist)} videos.")
    for vid in vidlist:
        print(f"  {vid.url}")
    return 0


if __name__ == "__main__":
    main()



"""
YouTube scraper.
"""

# pylint: disable=invalid-name,consider-using-f-string,fixme,missing-function-docstring,R0801

import datetime
import json
import os
import re
import subprocess
import sys
import tempfile
import time
from typing import Any, List, Optional

import feedparser  # type: ignore
import requests  # type: ignore
from bs4 import BeautifulSoup  # type: ignore
from keyvalue_sqlite import KeyValueSqlite as KeyValueDB  # type: ignore

from .date import iso8601_duration_as_seconds, iso_fmt, now_local
from .error import log_error
from .fetch_html import fetch_html, fetch_html_using_request_lib
from .video_info import VideoInfo

_ENABLE_PROFILE_FETCH = False

HERE = os.path.dirname(__file__)
DB_YOUTUBE_CACHE = os.path.join(HERE, "cache", "youtube_cache.db")
os.makedirs(os.path.dirname(DB_YOUTUBE_CACHE), exist_ok=True)

# Retired. Lots of cost and only 10% of videos have subtitles.
# subtitle_fetcher_cache = YtSubtitleFetcherCache(DB_YOUTUBE_SUBTITLE)


def _get_cache(cache_path: str) -> KeyValueDB:
    return KeyValueDB(cache_path, "cached_youtube_video_attributes")


def _try_get_cached_duration(url: str, cache_path: Optional[str]) -> Optional[int]:
    if cache_path is None:
        return None
    cache = _get_cache(cache_path)
    cached_data = cache.get(url)
    if cached_data:
        duration = cached_data.get("duration_seconds")
        if duration is not None:
            return int(duration)
    return None


def _set_cached_duration(url: str, duration_secs: int, cache_path: Optional[str]) -> None:
    if cache_path is None:
        return
    cache = _get_cache(cache_path)
    cached_data = cache.get(url)
    if not cached_data:
        cached_data = {}
    cached_data.update({"duration_seconds": duration_secs})
    cache[url] = cached_data


def strfdelta(duration_seconds: int) -> str:
    if duration_seconds == 0:
        return "Live"
    td = datetime.timedelta(seconds=duration_seconds)
    hours, rem = divmod(td.seconds, 3600)
    mins, seconds = divmod(rem, 60)
    if hours == 0:
        return "%d:%02d" % (mins, seconds)
    return "%d:%02d:%02d" % (hours, mins, seconds)


def fetch_youtube_duration_str(url: str, cache_path: Optional[str]) -> str:
    sys.stdout.write("  Youtube visiting video %s\n" % url)
    cached_duration = _try_get_cached_duration(url, cache_path)
    if cached_duration:
        if cached_duration > -1:
            return strfdelta(cached_duration)
        return ""  # Gracefully handle error condition.
    try:
        html_doc = fetch_html(url)
        assert html_doc, f"{__file__}: Could not fetch html doc from {url}"
        soup = BeautifulSoup(html_doc, "html.parser")
        dom = soup.find("meta", {"itemprop": "duration"})
        assert dom, f"{__file__}: Could not find duration in html doc in {url}"
        duration_str = dom.attrs["content"]
        duration_seconds = iso8601_duration_as_seconds(duration_str)
        _set_cached_duration(url, duration_seconds, cache_path)
        out = strfdelta(duration_seconds)
        return out
    except requests.exceptions.HTTPError as e:
        sys.stderr.write(f'{__file__} Error while processing {url} for duration because "{str(e)}"\n')
        _set_cached_duration(url, -1, cache_path)
        return ""


def _fetch_youtube_channel_via_rss(channel_name: str, channel_id: str) -> List[VideoInfo]:
    url = "https://www.youtube.com/feeds/videos.xml?channel_id=" + channel_id
    sys.stdout.write(f"Youtube visiting {channel_name} ({url})\n")
    content = fetch_html(url)
    if "was not found on this server" in content:  # TODO: Make less hacky.
        raise OSError(f"Could not fetch {url}")
    feed = feedparser.parse(content)
    output: List[VideoInfo] = []
    profile_picture = None
    for entry in feed.entries:
        views = entry.media_statistics["views"]
        if views == "0":  # Skip views with 0 as they have not be released yet.
            continue
        # Disable profile picture for now.
        if _ENABLE_PROFILE_FETCH and profile_picture is None:
            sys.stdout.write(f"  Youtube visiting {entry.link} to fetch profile picture.\n")
            vid_html = fetch_html(entry.link)
            if vid_html is None:
                sys.stdout.write(f"    Error - vid_html from {entry.link} was None")
            else:
                try:
                    vid_info = parse_youtube_video(vid_html)
                    profile_picture = vid_info.get("profile_thumbnail", None)
                except AttributeError as err:
                    sys.stdout.write(f"Failed to get photo from {entry.link} because {err}.\n")
        now_datestr = iso_fmt(now_local())
        o = VideoInfo(
            channel_name=channel_name,
            title=entry.title,
            date_published=entry.published,
            date_discovered=now_datestr,
            date_lastupdated=now_datestr,
            channel_url=entry.author_detail.href,
            source="youtube.com",
            url=entry.link,
            duration="?",
            description=entry.summary,
            img_src=entry.media_thumbnail[0]["url"],
            iframe_src="https://youtube.com/embed/" + entry.yt_videoid,
            views=views,
            profile_img_src=profile_picture or "",
        )
        output.append(o)
    return output


def _fetch_youtube_channel_via_html(
    channel_name: str,
    channel_id: str,
    limit: int = -1,
) -> List[Any]:
    """SLOW! Also, Experimental - use if the videos.xml api is disabled"""

    # Step 1: using the channel_id fetch the channel_name.
    # TODO: this should be cached to avoid a fetch.
    url = f"https://www.youtube.com/channel/{channel_id}"
    sys.stdout.write(f"Youtube visiting {channel_name} ({url})\n")
    html_doc = fetch_html(url)
    pat = r'\"canonicalBaseUrl\":\"/c/([^\"]+)"'
    canonical_name = re.findall(pat, html_doc)[0]
    # Now the fetch the videos list from the channels canonical name
    # url.
    url = f"https://youtube.com/c/{canonical_name}/videos"
    chan_html_doc = fetch_html_using_request_lib(url)
    # Inside a json file there is a bunch of videos that can be found
    # using the following regex.
    pat = r"{\"videoId\":\"([^\"]+)\","
    vid_ids = list(set(re.findall(pat, chan_html_doc)))
    vid_urls = [f"https://youtube.com/watch?v={id}" for id in vid_ids]
    output = []
    # Iterate through all videos and use the youtube-dl python command to
    # get a json dump of the file. Unfortunately this does not include the
    # very import publishing date time stamp. Another note is that with a
    # little bit of research it's not clear that there are alternative ways
    # to get exact timestamps from Youtube other than by videos.xml api, or
    # the youtube api which requires an audit of the app. Work arounds include
    # cross referencing the videos with those scraped from social blade or using
    # a paid service to get these videos.
    for i, vid_url in enumerate(vid_urls):
        if limit != -1 and i >= limit:
            break
        stdout = subprocess.check_output(f"youtube-dl {vid_url} -j", shell=True).decode("utf-8")
        data = json.loads(stdout)
        # Future:
        #  title = data['title']
        #  view_count = data['view_count']
        output.append(data)
    return output


def fetch_youtube_today(
    channel_name: str,
    channel_id: str,
    cache_path: Optional[str] = DB_YOUTUBE_CACHE,
    limit: int = -1,
) -> List[VideoInfo]:
    start_time = time.time()
    video_list = _fetch_youtube_channel_via_rss(channel_name, channel_id)
    if limit != -1:
        video_list = video_list[0:limit]
    vid: VideoInfo
    for vid in video_list:
        # Add in duration with per-video fetch
        duration_time = fetch_youtube_duration_str(vid.url, cache_path)
        vid.duration = duration_time
    delta_time = start_time - time.time()
    if delta_time > 15:
        sys.stdout.write(f"WARNING, youtube scraper took {int(delta_time)} seconds to complete\n")
    return video_list


def _test_fetch_youtube_video_info(url: str) -> None:
    sys.stdout.write("Youtube visiting video %s\n" % url)
    html_doc = fetch_html(url)
    premiered_ago_matches = re.findall(r'\"Premiered (.{1,15}) ago"}', html_doc)
    premiered_in_progress = re.findall(r'\"Premiere in progress\. Started (.{1,15}) ago"}', html_doc)
    started_streaming_matches = re.findall(r"\"Started streaming (.{1,15}) ago\"", html_doc)
    video_is_private = '{"simpleText":"Private video"}' in html_doc
    unplayable = '"status":"UNPLAYABLE"' in html_doc
    # has_set_reminder = re.findall(r'\"simpleText\":\"Set reminder\"', html_doc)
    # has_set_reminder = '{"simpleText":"Set reminder"}' in html_doc

    stream_offline = '"status":"LIVE_STREAM_OFFLINE"' in html_doc

    print("  Testing %s" % url)
    print("    Premiered ago: %s" % premiered_ago_matches)
    print("    Premiered in progress started: %s" % premiered_in_progress)
    print("    Started streaming ago: %s" % started_streaming_matches)
    print("    Has stream_offline: %s" % stream_offline)
    print("    Is unplayable: %s" % unplayable)
    print("    Video is private: %s" % video_is_private)


# Still experimental
def parse_youtube_video(html_doc: str) -> dict:
    out: dict = {
        "date_published": "?",
        "views": "?",
        "profile_thumbnail": "?",
        "is_live": "?",
    }
    if '"isLiveNow":true' in html_doc:
        out["is_live"] = "True"
    else:
        out["is_live"] = "False"
    top_dom = BeautifulSoup(html_doc, "html.parser")  # type: ignore
    try:
        dom = top_dom.find("meta", {"itemprop": "startDate"})  # type: ignore
        out["date_published"] = str(dom.attrs["content"])
    except KeyError as err:
        log_error(str(err))
    except AttributeError as err:
        text = "NO STATUS"
        try:
            text = top_dom.text
        except Exception as err2:  # pylint: disable=broad-except
            print(f"{__file__}: Error: {err2}")
        sys.stdout.write(f"{__file__}: \ntop_dom.text: {text}\n\nError: {err}\n")
        raise
    try:
        dom = top_dom.find("meta", {"itemprop": "interactionCount"})  # type: ignore
        out["views"] = str(dom.attrs["content"])
    except KeyError as ke:
        log_error(str(ke))
    all_scripts = top_dom.find_all("script")
    needle_initial_data = "var ytInitialData = "
    # Needed for commented out code, below.
    # needle_player_response = "var ytInitialPlayerResponse = "
    for script in all_scripts:
        if not script.string:
            continue
        try:
            if script.string.startswith(needle_initial_data):
                img_urls = re.findall(r"\"(https://yt\d[^\"]+)\"", script.string)
                for img_url in img_urls:
                    if "s176" in img_url:  # For some reason this appears only in the thumbnail image.
                        out["profile_thumbnail"] = img_url
                        break
            # Video Details contains a lot of gems, but is disabled for now.
            # if script.string.startswith(needle_player_response):
            #     last_semi: int = script.string.rfind(";")
            #     json_str = script.string[
            #         len(needle_player_response) : last_semi
            #     ]
            #     json_data = json.loads(json_str)
            #     video_details = json_data["videoDetails"]
        except BaseException as be:  # pylint: disable=broad-except
            log_error("Error: " + str(be) + " for " + str(out))
    return out


def test_fetch_duration():
    url = "https://www.youtube.com/watch?v=UywxSWjzGaU"
    try:
        cache_path_file = tempfile.NamedTemporaryFile(suffix=".db", delete=False)  # pylint: disable=consider-using-with
        cache_path = cache_path_file.name
        cache_path_file.close()
        duration_str = fetch_youtube_duration_str(url, cache_path=cache_path)
        print("Discovered that the video at %s is %s" % (url, duration_str))
        duration_str = fetch_youtube_duration_str(url, cache_path=cache_path)
        print("Discovered that the video at %s is %s" % (url, duration_str))
    finally:
        os.remove(cache_path)


if __name__ == "__main__":
    # test_fetch_duration()
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=5nBqIK0mrFI")
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=r5VjrVqgEU8")
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=xfJsiDkbjg0")
    # Project veritas should not return anything
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=F7y19uNaYfE")
    # Upcoming film
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=yv7NFn95R0s")
    _test_fetch_youtube_video_info("https://www.youtube.com/watch?v=egRMAnwUQDQ")



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring
import unittest
from datetime import datetime
from typing import Optional

from vidcrawler.date import now_local, parse_datetime
from vidcrawler.rumble import fetch_rumble, fetch_rumble_channel_today
from vidcrawler.video_info import VideoInfo


class RumbleScraperTester(unittest.TestCase):
    def test_fetch_rumble(self):
        vid_list = fetch_rumble(channel="DiamondandSilk")
        self.assertTrue(vid_list)

    def test_fetch_alt_channel_url(self):
        vid_list = fetch_rumble(channel="MaryamXHenein")
        self.assertTrue(vid_list)

    def test_fetch_bannon(self) -> None:
        max_days = 4
        vid_list = fetch_rumble_channel_today(channel="BannonsWarRoom", channel_name="BannonsWarRoom")
        most_recent_date: Optional[datetime] = None
        vid: VideoInfo
        for vid in vid_list:
            if most_recent_date is None:
                most_recent_date = parse_datetime(vid.date_published)
                continue
            new_datetime = parse_datetime(vid.date_published)
            if new_datetime > most_recent_date:
                most_recent_date = new_datetime
        # get time delta between most recent date and now in days
        bannons_last_show_in_days: float = 9999
        if most_recent_date is not None:
            bannons_last_show_in_days = (now_local() - most_recent_date).total_seconds() / (3600 * 24)

        self.assertLess(
            bannons_last_show_in_days,
            max_days,
            f"Bannon's last show was {bannons_last_show_in_days} days ago.",
        )


if __name__ == "__main__":
    unittest.main()



"""
Command entry point.
"""

# pylint: disable=consider-using-f-string

import argparse
import json
import os
import subprocess
import warnings

from filelock import FileLock
from static_ffmpeg import add_paths

from vidcrawler.youtube_bot import YtVid, fetch_all_vids


def parse_args() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser("youtube-pull")
    parser.add_argument(
        "channel",
        type=str,
        # help="URL of the channel, example: https://www.youtube.com/@silverguru/videos",
        help="URL slug of the channel, example: @silverguru",
    )
    parser.add_argument("basedir", type=str)
    parser.add_argument(
        "--limit-scroll-pages",
        type=int,
        default=1000,
        help="Limit the number of the number of pages to scroll down",
    )
    parser.add_argument(
        "--download",
        action="store_true",
        help="Deprecated: This option does nothing.",
    )
    parser.add_argument(
        "--skip-download",
        action="store_true",
        help="Skip the download of the videos.",
    )
    parser.add_argument(
        "--download-limit",
        type=int,
        default=-1,
        help="Limit the number of videos to download",
    )
    parser.add_argument(
        "--skip-scan",
        action="store_true",
        help="Skip the update of the library.json file",
    )
    return parser.parse_args()


def load_json(file_path: str) -> list[dict]:
    """Load json from file if it exists."""
    if not os.path.exists(file_path):
        return []
    with open(file_path, encoding="utf-8", mode="r") as filed:
        # data = json.load(filed)
        # return data
        json_str = filed.read()
    json_obj = json.loads(json_str)
    return json_obj


def save_json(file_path: str, data: list[dict]) -> None:
    """Save json to file."""
    out_str = json.dumps(data, indent=2)
    with open(file_path, encoding="utf-8", mode="w") as filed:
        filed.write(out_str)


def to_channel_url(channel: str) -> str:
    """Convert channel name to channel URL."""
    out = f"https://www.youtube.com/{channel}/videos"
    return out


def yt_dlp_download_mp3(url: str, outmp3: str) -> None:
    """Download the youtube video as an mp3."""
    add_paths()
    par_dir = os.path.dirname(outmp3)
    if par_dir:
        os.makedirs(par_dir, exist_ok=True)

    for _ in range(3):
        try:
            cmd_list: list[str] = [
                "yt-dlp",
                url,
                "--extract-audio",
                "--audio-format",
                "mp3",
                "--output",
                outmp3,
            ]
            subprocess.run(cmd_list, check=True)
            return
        except subprocess.CalledProcessError as cpe:
            print(f"Failed to download {url} as mp3: {cpe}")
            continue
    warnings.warn(f"Failed all attempts to download {url} as mp3.")


def find_missing_downloads(library_json_path: str) -> list[dict]:
    """Find missing downloads."""
    out: list[dict] = []
    with open(library_json_path, encoding="utf-8", mode="r") as filed:
        data = json.load(filed)
    for vid in data:
        title = vid["title"]
        file_path = os.path.join(os.path.dirname(library_json_path), f"{title}.mp3")
        if not os.path.exists(file_path):
            vid["file_path"] = file_path
            out.append(vid)
    return out


def main() -> None:
    """Main function."""
    args = parse_args()
    channel_url = to_channel_url(args.channel)
    output_dir = os.path.join(args.basedir, args.channel, "youtube")
    limit_scroll_pages = args.limit_scroll_pages
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    library_json = os.path.join(output_dir, "library.json")
    # load the json data if it already exists
    file_lock = library_json + ".lock"
    if not args.skip_scan:
        with FileLock(file_lock):
            loaded_data = load_json(library_json)
            vids: list[YtVid] = fetch_all_vids(channel_url, limit=limit_scroll_pages)
            fetched_data = [vid.to_dict() for vid in vids]
            new_data = list(loaded_data)
            for vid in fetched_data:
                if vid not in loaded_data:
                    new_data.append(vid)
            save_json(library_json, new_data)
        print(f"Updated {library_json}")

    if not os.path.exists(library_json):
        raise FileNotFoundError(f"{library_json} does not exist. Cannot skip scan.")
    if args.download:
        print("Warning: The --download option is deprecated is now implied. Use --skip-download to avoid downloading")
    if not args.skip_download:
        download_count = 0
        while True:
            if args.download_limit != -1 and download_count >= args.download_limit:
                break
            missing_downloads = find_missing_downloads(library_json)
            if not missing_downloads:
                break
            vid = missing_downloads[0]
            next_url = vid["url"]
            next_mp3_path = vid["file_path"]
            print(f"\n#######################\n# Downloading missing file {next_url}: {next_mp3_path}\n" "###################")
            yt_dlp_download_mp3(url=next_url, outmp3=next_mp3_path)
            download_count += 1


if __name__ == "__main__":
    import sys

    sys.argv.append("@silverguru")
    sys.argv.append("tmp")
    sys.argv.append("--limit-scroll-pages")
    sys.argv.append("1")
    main()



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import unittest

from vidcrawler.brighteon import fetch_brighteon_today


class BrighteonScraperTester(unittest.TestCase):
    def test_fetch_brighteon_today(self):
        vid_list = fetch_brighteon_today(channel_name="brendon", channel="brendon")
        self.assertIsNotNone(vid_list)


if __name__ == "__main__":
    unittest.main()



"""
Tests bitchute scraper.
"""

# pylint: disable=missing-class-docstring,missing-function-docstring


import os
import unittest

from vidcrawler.bitchute import fetch_bitchute_today, fetch_rss_url, parse_rss_feed
from vidcrawler.fetch_html import fetch_html

IS_GITHUB_RUNNER = (
    "/home/runner" in os.environ.get("TOX_ENV_DIR", "")
    or "D:\\a\\vidcrawler" in os.environ.get("TOX_PACKAGE", "")
    or "/Users/runner/" in os.environ.get("TOX_WORK_DIR", "")
    or "/Users/runner/" in os.environ.get("VIRTUAL_ENV", "")
    or "/Users/runner/" in os.environ.get("PATH", "")
)


class BitchuteScraperTester(unittest.TestCase):
    """Tester for bitchute."""

    @unittest.skipIf(IS_GITHUB_RUNNER, "Skip amazing polly on github actions, it fails.")
    def test_fetch_rss_video(self):
        rss_url = fetch_rss_url(channel_name="Infowars", channel_id="9c7qJvwx7YQT")
        self.assertIsNotNone(rss_url)
        self.assertEqual(
            "https://www.bitchute.com/feeds/rss/channel/banned-dot-video/",
            rss_url,
        )
        content: str = fetch_html(rss_url)  # type: ignore
        feed = parse_rss_feed(content)
        self.assertIsNotNone(feed)

    @unittest.skipIf(IS_GITHUB_RUNNER, "Skip amazing polly on github actions, it fails.")
    def test_fetch_bitchute_today(self):
        vid_list = fetch_bitchute_today(channel_name="Infowars", channel_id="9c7qJvwx7YQT")
        self.assertIsNotNone(vid_list)

    @unittest.skipIf(IS_GITHUB_RUNNER, "Skip amazing polly on github actions, it fails.")
    def test_fetch_bitchute_amazing_polly(self):
        vid_list = fetch_bitchute_today(channel_name="Amazing Polly", channel_id="ZofFQQoDoqYT")
        self.assertTrue(vid_list)
        vid = vid_list[0]
        self.assertEqual("Amazing Polly", vid.channel_name)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-module-docstring,missing-class-docstring,missing-function-docstring

import os
import re
import unittest
from typing import List

from vidcrawler.odysee import fetch_odysee_today
from vidcrawler.video_info import VideoInfo

RUN_ALL_TESTS = os.environ.get("RUN_ALL_TESTS", None) is not None


def is_valid_url(url: str) -> bool:
    # Create a regex to match a url
    regex = r"^(?:http(s)?:\/\/)?[\w.-]+(?:\.[\w\.-]+)+[\w\-\._~:/?#[\]@!\$&'\(\)\*\+,;=.]+$"
    return re.match(regex, url) is not None


class OdyseeScraperTester(unittest.TestCase):
    @unittest.skipUnless(RUN_ALL_TESTS, "Odysee is not working")
    def test_fetch_odysee_today(self) -> None:
        # RSS FEED:
        #   https://lbryfeed.melroy.org/channel/odysee/BretWeinstein
        vid_list: List[VideoInfo] = fetch_odysee_today(channel_name="BretWeinstein", channel="BretWeinstein")
        self.assertGreater(len(vid_list), 0)

    @unittest.skipUnless(RUN_ALL_TESTS, "Odysee is not working")
    def test_union_of_the_unwanted(self) -> None:
        vid_list: List[VideoInfo] = fetch_odysee_today(channel_name="UOTU", channel="uotuw:e")
        self.assertGreater(len(vid_list), 0)

    @unittest.skipUnless(RUN_ALL_TESTS, "Odysee is not working")
    def test_odysee_bug(self) -> None:
        # Test the fix for odysee bug:
        #   https://github.com/zackees/blast.video/issues/1
        vid_list: List[VideoInfo] = fetch_odysee_today(channel_name="The Ripple Effect", channel="therippleeffectpodcast")
        self.assertGreater(len(vid_list), 0)
        for vid in vid_list:
            self.assertEqual(vid.channel_name, "The Ripple Effect")
            self.assertTrue(is_valid_url(vid.img_src), f'"{vid.img_src}" is not a valid url')


if __name__ == "__main__":
    unittest.main()



"""
    Scraper for Brighteon.com
"""

# pylint: disable=line-too-long,missing-function-docstring,consider-using-f-string,too-many-locals,invalid-name,no-else-return


import json
import sys
from typing import List

from bs4 import BeautifulSoup  # type: ignore

from .date import iso_fmt, now_local
from .fetch_html import fetch_html_using_request_lib as fetch_html
from .video_info import VideoInfo


def fetch_brighteon_today(channel_name: str, channel: str) -> List[VideoInfo]:
    output: List[VideoInfo] = []
    url = "https://www.brighteon.com/channels/%s" % channel
    sys.stdout.write("Brighteon visiting %s (%s)\n" % (channel, url))
    html_doc = fetch_html(url)
    soup = BeautifulSoup(html_doc, "html.parser")
    json_data = soup.find("script", id="__NEXT_DATA__").string
    data = json.loads(json_data)
    channel_data = data["props"]["initialProps"]["pageProps"]["data"]
    videos = channel_data["videos"]
    for vid in videos:
        created_time = vid["createdAt"]
        id = vid["id"]  # pylint: disable=redefined-builtin
        iframe_url = "https://www.brighteon.com/embed%s" % id
        video_url = "https://www.brighteon.com/%s" % id
        title = vid["name"]
        image_src = vid["thumbnail"]
        view_count = vid["analytics"]["videoView"]
        duration = vid["duration"]
        now_datestr = iso_fmt(now_local())
        o = VideoInfo(
            channel_name=channel_name,
            source="brighteon.com",
            date_published=created_time,
            date_discovered=now_datestr,
            date_lastupdated=now_datestr,
            url=video_url,
            channel_url=url,
            title=title,
            duration=duration,
            description="TODO",
            img_src=image_src,
            iframe_src=iframe_url,
            views=view_count,
            profile_img_src="",
        )
        output.append(o)
    return output
